import tensorflow as tf


class FeedForward(tf.keras.layers.Layer):
    def __init__(
        self,
        output_units,
        hidden_units,
        dropout_rate=0.1,
        residual_smoothing=True,
        name=None,
        dtype=None,
    ):
        super().__init__(name=name, dtype=dtype)
        self._output_units = int(output_units)
        self._hidden_units = int(hidden_units)
        self._dropout_rate = float(dropout_rate)
        self._residual_smoothing = bool(residual_smoothing)

        if self._residual_smoothing:
            self._emb_layer = tf.keras.layers.Dense(
                self._output_units, dtype=self.dtype
            )
        else:
            self._emb_layer = None

        self._seq = tf.keras.Sequential(
            [
                tf.keras.layers.Dense(
                    self._hidden_units, activation="relu", dtype=self.dtype
                ),
                tf.keras.layers.Dropout(self._dropout_rate, dtype=self.dtype),
                tf.keras.layers.Dense(self._output_units, dtype=self.dtype),
            ]
        )

        self._add = tf.keras.layers.Add()
        self._layer_norm = tf.keras.layers.LayerNormalization(dtype=self.dtype)

    def call(self, x):
        if self._emb_layer is not None:
            x = self._emb_layer(x)
        x = self._add([x, self._seq(x)])
        x = self._layer_norm(x)
        return x  # shape (batch_size, x_elements, output_units)

    @property
    def output_units(self) -> int:
        return self._output_units

    @property
    def hidden_units(self) -> int:
        return self._hidden_units

    @property
    def dropout_rate(self) -> float:
        return self._dropout_rate

    @property
    def residual_smoothing(self) -> bool:
        return self._residual_smoothing



from .BinaryCrossentropy import BinaryCrossentropy
from .CaloLoss import CaloLoss
from .JSDivergence import JSDivergence
from .KLDivergence import KLDivergence
from .MeanAbsoluteError import MeanAbsoluteError
from .MeanSquaredError import MeanSquaredError



import tensorflow as tf

from calotron.utils import checkActivations


class MultiActivations(tf.keras.layers.Layer):
    def __init__(self, activations, output_depth, name=None, dtype=None):
        super().__init__(name=name, dtype=dtype)
        self._output_depth = int(output_depth)
        self._output_activations = checkActivations(activations, output_depth, dtype)

    def call(self, x):
        if x.shape[2] != self._output_depth:
            raise ValueError(
                f"`output_depth` passed {self._output_depth} "
                f"doesn't match with the input tensor shape "
                f"({x.shape})."
            )
        if self._output_activations is not None:
            concat = list()
            for i, activation in enumerate(self._output_activations):
                concat.append(activation(x[:, :, i])[:, :, None])
            x = tf.concat(concat, axis=2)
        return x  # shape (batch_size, x_elements, x_depth)

    @property
    def output_activations(self):  # TODO: add Union[list, None]
        return self._output_activations

    @property
    def output_depth(self) -> int:
        return self._output_depth



import tensorflow as tf

from calotron.layers import DeepSets


class Discriminator(tf.keras.Model):
    def __init__(
        self,
        latent_dim,
        output_units,
        output_activation=None,
        hidden_layers=3,
        hidden_units=128,
        dropout_rate=0.1,
        name=None,
        dtype=None,
    ):
        super().__init__(name=name, dtype=dtype)
        self._latent_dim = int(latent_dim)
        self._output_units = int(output_units)
        self._output_activation = output_activation
        self._hidden_layers = int(hidden_layers)
        self._hidden_units = int(hidden_units)
        self._dropout_rate = float(dropout_rate)

        self._deepsets = DeepSets(
            latent_dim=self._latent_dim,
            num_layers=self._hidden_layers,
            hidden_units=self._hidden_units,
            dropout_rate=self._dropout_rate,
            dtype=self.dtype,
        )

        self._seq = [
            tf.keras.layers.Dense(
                self._latent_dim, activation="relu", dtype=self.dtype
            ),
            tf.keras.layers.Dropout(self._dropout_rate, dtype=self.dtype),
            tf.keras.layers.Dense(
                self._latent_dim, activation="relu", dtype=self.dtype
            ),
            tf.keras.layers.Dropout(self._dropout_rate, dtype=self.dtype),
        ]
        self._seq += [
            tf.keras.layers.Dense(
                self._output_units, activation=self._output_activation, dtype=self.dtype
            )
        ]

    def call(self, x):
        x = self._deepsets(x)
        for layer in self._seq:
            x = layer(x)
        return x

    @property
    def latent_dim(self) -> int:
        return self._latent_dim

    @property
    def output_units(self) -> int:
        return self._output_units

    @property
    def output_activation(self):  # TODO: add Union[None, activation]
        return self._output_activation

    @property
    def hidden_layers(self) -> int:
        return self._hidden_layers

    @property
    def hidden_units(self) -> int:
        return self._hidden_units

    @property
    def dropout_rate(self) -> float:
        return self._dropout_rate

    @property
    def deepsets(self) -> DeepSets:
        return self._deepsets



import tensorflow as tf
from tensorflow.keras.losses import BinaryCrossentropy as TF_BCE

from calotron.losses.BaseLoss import BaseLoss


class BinaryCrossentropy(BaseLoss):
    def __init__(
        self,
        from_logits=False,
        label_smoothing=0.0,
        axis=-1,
        reduction="auto",
        name="bce_loss",
    ):
        super().__init__(name)
        self._loss = TF_BCE(
            from_logits=from_logits,
            label_smoothing=label_smoothing,
            axis=axis,
            reduction=reduction,
        )

    def discriminator_loss(
        self, discriminator, target_true, target_pred, sample_weight=None
    ):
        y_true = discriminator(target_true)
        y_pred = discriminator(target_pred)
        loss_real = self._loss(
            tf.ones_like(y_true), y_true, sample_weight=sample_weight
        )
        loss_fake = self._loss(
            tf.zeros_like(y_pred), y_pred, sample_weight=sample_weight
        )
        return loss_real + loss_fake

    def transformer_loss(
        self, discriminator, target_true, target_pred, sample_weight=None
    ):
        y_pred = discriminator(target_pred)
        loss_fake = self._loss(
            tf.ones_like(y_pred), y_pred, sample_weight=sample_weight
        )
        return loss_fake



import tensorflow as tf
from tensorflow.keras.losses import BinaryCrossentropy as TF_BCE
from tensorflow.keras.losses import MeanSquaredError as TF_MSE

from calotron.losses.BaseLoss import BaseLoss


class CaloLoss(BaseLoss):
    def __init__(
        self,
        alpha=0.1,
        from_logits=False,
        label_smoothing=0.0,
        axis=-1,
        reduction="auto",
        name="calo_loss",
    ):
        super().__init__(name)
        self._alpha = float(alpha)
        self._mse_loss = TF_MSE(reduction=reduction)
        self._bce_loss = TF_BCE(
            from_logits=from_logits,
            label_smoothing=label_smoothing,
            axis=axis,
            reduction=reduction,
        )

    def discriminator_loss(
        self, discriminator, target_true, target_pred, sample_weight=None
    ):
        y_true = discriminator(target_true)
        y_pred = discriminator(target_pred)
        loss_real = self._bce_loss(
            tf.ones_like(y_true), y_true, sample_weight=sample_weight
        )
        loss_fake = self._bce_loss(
            tf.zeros_like(y_pred), y_pred, sample_weight=sample_weight
        )
        return loss_real + loss_fake

    def transformer_loss(
        self, discriminator, target_true, target_pred, sample_weight=None
    ):
        mse_loss = self._mse_loss(target_true, target_pred, sample_weight=sample_weight)
        y_pred = discriminator(target_pred)
        bce_loss = self._bce_loss(
            tf.ones_like(y_pred), y_pred, sample_weight=sample_weight
        )
        return mse_loss + self._alpha * bce_loss



import numpy as np
import tensorflow as tf


class PositionalEmbedding(tf.keras.layers.Layer):
    def __init__(
        self,
        output_depth,
        max_length=32,
        encoding_normalization=128,
        dropout_rate=0.1,
        name=None,
        dtype=None,
    ):
        super().__init__(name=name, dtype=dtype)
        self._output_depth = int(output_depth)
        self._max_length = int(max_length)
        self._encoding_normalization = float(encoding_normalization)
        self._dropout_rate = float(dropout_rate)

        self._embedding = tf.keras.layers.Dense(
            self._output_depth, activation="linear", dtype=self.dtype
        )

        self._pos_encoding = self._positional_encoding(
            length=self._max_length,
            depth=self._output_depth,
            normalization=self._encoding_normalization,
            dtype=self.dtype,
        )

        self._dropout = tf.keras.layers.Dropout(self._dropout_rate, dtype=self.dtype)

    def call(self, x):
        length = tf.shape(x)[1]
        x = self._embedding(x)
        x *= tf.math.sqrt(tf.cast(self._output_depth, self.dtype))  # scale factor
        x = x + self._pos_encoding[None, :length, :]
        x = self._dropout(x)
        return x

    @staticmethod
    def _positional_encoding(length, depth, normalization=512, dtype=tf.float32):
        pos_encoding = np.zeros(shape=(length, depth))  # buffer to fill
        for k in range(length):
            for i in range(int(depth / 2)):
                denominator = np.power(normalization, 2 * i / depth)
                pos_encoding[k, 2 * i] = np.sin(k / denominator)
                pos_encoding[k, 2 * i + 1] = np.cos(k / denominator)
        return tf.cast(pos_encoding, dtype=dtype)

    @property
    def output_depth(self) -> int:
        return self._output_depth

    @property
    def max_length(self) -> int:
        return self._max_length

    @property
    def encoding_normalization(self) -> float:
        return self._encoding_normalization

    @property
    def dropout_rate(self) -> float:
        return self._dropout_rate



import tensorflow as tf


class DeepSets(tf.keras.layers.Layer):
    def __init__(
        self,
        latent_dim,
        num_layers,
        hidden_units=256,
        dropout_rate=0.1,
        name=None,
        dtype=None,
    ):
        super().__init__(name=name, dtype=dtype)
        self._latent_dim = int(latent_dim)
        if num_layers < 1:
            raise ValueError("`num_layers` should be greater than 0")
        self._num_layers = int(num_layers)
        self._hidden_units = int(hidden_units)
        self._dropout_rate = float(dropout_rate)

        self._seq = list()
        for _ in range(self._num_layers - 1):
            self._seq.append(
                tf.keras.layers.Dense(
                    self._hidden_units, activation="relu", dtype=self.dtype
                )
            )
            self._seq.append(
                tf.keras.layers.Dropout(self._dropout_rate, dtype=self.dtype)
            )
        self._seq += [
            tf.keras.layers.Dense(self._latent_dim, activation="relu", dtype=self.dtype)
        ]

    def call(self, x):
        # shape: (batch_size, x_elements, x_depth)
        outputs = list()
        for i in range(
            x.shape[1]
        ):  # TODO: redesign computation (parallel instead serial)
            latent_tensor = x[:, i : i + 1, :]  # shape: (batch_size, 1, x_depth)
            for layer in self._seq:
                latent_tensor = layer(
                    latent_tensor
                )  # shape: (batch_size, 1, latent_dim)
            outputs.append(latent_tensor)

        concat = tf.keras.layers.Concatenate(axis=1)(
            outputs
        )  # shape: (batch_size, x_elements, latent_dim)
        output = tf.reduce_sum(concat, axis=1)  # shape: (batch_size, latent_dim)
        return output

    @property
    def latent_dim(self) -> int:
        return self._latent_dim

    @property
    def num_layers(self) -> int:
        return self._latent_dim

    @property
    def hidden_units(self) -> int:
        return self._hidden_units

    @property
    def dropout_rate(self) -> float:
        return self._dropout_rate



import tensorflow as tf

from calotron.models.Calotron import Calotron
from calotron.models.Generator import Generator


class Calogantron(Calotron):
    def __init__(self, transformer, discriminator, generator):
        super().__init__(transformer, discriminator)
        if not isinstance(generator, Generator):
            raise TypeError(
                f"`generator` should be a calotron's "
                f"`Generator`, instead "
                f"{type(generator)} passed"
            )
        self._generator = generator

    def call(self, inputs):
        source, target = inputs
        output, d_output_true, d_output_pred = super().call(inputs)
        start_token = self.get_start_token(target)
        return output, d_output_true, d_output_pred, start_token

    def summary(self):
        self._transformer.summary()
        self._discriminator.summary()
        self._generator.summary()

    @property
    def generator(self) -> Generator:
        return self._generator



import tensorflow as tf
from tensorflow.keras.metrics import MeanAbsoluteError as TF_MAE

from calotron.metrics.BaseMetric import BaseMetric


class MeanAbsoluteError(BaseMetric):
    def __init__(self, name="mae", dtype=None, **kwargs):
        super().__init__(name, dtype, **kwargs)
        self._mae = TF_MAE(name=name, dtype=dtype)

    def update_state(self, y_true, y_pred, sample_weight=None):
        state = self._mae(y_true, y_pred, sample_weight=sample_weight)
        self._metric_values.assign(state)



import tensorflow as tf
from tensorflow.keras.losses import MeanAbsoluteError as TF_MAE

from calotron.losses.BaseLoss import BaseLoss


class MeanAbsoluteError(BaseLoss):
    def __init__(self, reduction="auto", name="mae_loss"):
        super().__init__(name)
        self._loss = TF_MAE(reduction=reduction)

    def discriminator_loss(
        self, discriminator, target_true, target_pred, sample_weight=None
    ):
        y_true = discriminator(target_true)
        y_pred = discriminator(target_pred)
        return -self._loss(
            y_true, y_pred, sample_weight=sample_weight
        )  # error maximization

    def transformer_loss(
        self, discriminator, target_true, target_pred, sample_weight=None
    ):
        y_true = discriminator(target_true)
        y_pred = discriminator(target_pred)
        return self._loss(
            y_true, y_pred, sample_weight=sample_weight
        )  # error minimization



import tensorflow as tf
from tensorflow.keras.metrics import BinaryAccuracy as TF_BinaryAccuracy

from calotron.metrics.BaseMetric import BaseMetric


class Accuracy(BaseMetric):
    def __init__(self, name="accuracy", dtype=None, threshold=0.5):
        super().__init__(name, dtype)
        self._accuracy = TF_BinaryAccuracy(name=name, dtype=dtype, threshold=threshold)

    def update_state(self, y_true, y_pred, sample_weight=None):
        state = self._accuracy(
            tf.ones_like(y_pred), y_pred, sample_weight=sample_weight
        )
        self._metric_values.assign(state)



import tensorflow as tf

from calotron.layers.Attention import GlobalSelfAttention
from calotron.layers.FeedForward import FeedForward
from calotron.layers.PositionalEmbedding import PositionalEmbedding


class EncoderLayer(tf.keras.layers.Layer):
    def __init__(
        self,
        encoder_depth,
        num_heads,
        key_dim=None,
        ff_units=256,
        dropout_rate=0.1,
        residual_smoothing=True,
        name=None,
        dtype=None,
    ):
        super().__init__(name=name, dtype=dtype)
        self._encoder_depth = int(encoder_depth)
        self._num_heads = int(num_heads)
        self._key_dim = int(key_dim) if key_dim else None
        self._ff_units = int(ff_units)
        self._dropout_rate = float(dropout_rate)
        self._residual_smoothing = bool(residual_smoothing)

        self._gsa_layer = GlobalSelfAttention(
            num_heads=self._num_heads,
            key_dim=self._key_dim if self._key_dim else self._encoder_depth,
            dropout=self._dropout_rate,
            dtype=self.dtype,
        )

        self._ff_layer = FeedForward(
            output_units=self._encoder_depth,
            hidden_units=self._ff_units,
            residual_smoothing=self._residual_smoothing,
            dtype=self.dtype,
        )

    def call(self, x):
        x = self._gsa_layer(x)  # shape: (batch_size, x_elements, x_depth)
        x = self._ff_layer(x)  # shape: (batch_size, x_elements, encoder_depth)
        return x

    @property
    def encoder_depth(self) -> int:
        return self._encoder_depth

    @property
    def num_heads(self) -> int:
        return self._num_heads

    @property
    def key_dim(self):  # TODO: add Union[int, None]
        return self._key_dim

    @property
    def ff_units(self) -> int:
        return self._ff_units

    @property
    def dropout_rate(self) -> float:
        return self._dropout_rate

    @property
    def residual_smoothing(self) -> bool:
        return self._residual_smoothing


class Encoder(tf.keras.layers.Layer):
    def __init__(
        self,
        encoder_depth,
        num_layers,
        num_heads,
        key_dim=None,
        pos_dim=None,
        pos_normalization=128,
        max_length=32,
        ff_units=256,
        dropout_rate=0.1,
        pos_sensitive=False,
        residual_smoothing=True,
        name=None,
        dtype=None,
    ):
        super().__init__(name=name, dtype=dtype)
        self._encoder_depth = int(encoder_depth)
        self._num_layers = int(num_layers)
        self._num_heads = int(num_heads)
        self._key_dim = int(key_dim) if key_dim else None
        self._pos_dim = int(pos_dim) if pos_dim else None
        self._pos_normalization = float(pos_normalization)
        self._max_length = int(max_length)
        self._ff_units = int(ff_units)
        self._dropout_rate = float(dropout_rate)
        self._pos_sensitive = bool(pos_sensitive)
        self._residual_smoothing = bool(residual_smoothing)

        if self._pos_sensitive:
            self._pos_embedding = PositionalEmbedding(
                self._pos_dim if self._pos_dim else self._encoder_depth,
                max_length=self._max_length,
                encoding_normalization=self._pos_normalization,
                dropout_rate=self._dropout_rate,
                dtype=self.dtype,
            )
        else:
            self._pos_embedding = None

        self._enc_layers = [
            EncoderLayer(
                encoder_depth=self._encoder_depth,
                num_heads=self._num_heads,
                key_dim=self._key_dim,
                ff_units=self._ff_units,
                dropout_rate=self._dropout_rate,
                residual_smoothing=self._residual_smoothing,
                dtype=self.dtype,
            )
            for _ in range(self._num_layers)
        ]

    def call(self, x):
        if self._pos_embedding is not None:
            x = self._pos_embedding(x)  # shape: (batch_size, x_elements, pos_dim)
        for i in range(self._num_layers):
            x = self._enc_layers[i](x)
        return x  # shape: (batch_size, x_elements, encoder_depth)

    @property
    def encoder_depth(self) -> int:
        return self._encoder_depth

    @property
    def num_layers(self) -> int:
        return self._num_layers

    @property
    def num_heads(self) -> int:
        return self._num_heads

    @property
    def key_dim(self):  # TODO: add Union[int, None]
        return self._key_dim

    @property
    def pos_dim(self):  # TODO: add Union[int, None]
        return self._pos_dim

    @property
    def pos_normalization(self) -> float:
        return self._pos_normalization

    @property
    def max_length(self) -> int:
        return self._max_length

    @property
    def ff_units(self) -> int:
        return self._ff_units

    @property
    def dropout_rate(self) -> float:
        return self._dropout_rate

    @property
    def pos_sensitive(self) -> bool:
        return self._pos_sensitive

    @property
    def residual_smoothing(self) -> bool:
        return self._residual_smoothing



import tensorflow as tf

from calotron.simulators.Simulator import Simulator

TF_FLOAT = tf.float32


class ExportSimulator(tf.Module):
    def __init__(self, simulator):
        super().__init__()
        if not isinstance(simulator, Simulator):
            raise TypeError(
                f"`transformer` should be a calotron's "
                f"`{type(Simulator).__name__}`, instead "
                f"{type(simulator)} passed"
            )
        self._simulator = simulator

    @tf.function(input_signature=[tf.TensorSpec(shape=[], dtype=TF_FLOAT)])
    def __call__(self, source, max_length):
        source = tf.cast(source, dtype=TF_FLOAT)
        result = self._simulator(source, max_length)
        return result

    @property
    def simulator(self) -> Simulator:
        return self._simulator



import tensorflow as tf
from tensorflow.keras.metrics import KLDivergence as TF_KLDivergence

from calotron.metrics.BaseMetric import BaseMetric


class JSDivergence(BaseMetric):
    def __init__(self, name="js_div", dtype=None):
        super().__init__(name, dtype)
        self._kl_div = TF_KLDivergence(name=name, dtype=dtype)

    def update_state(self, y_true, y_pred, sample_weight=None):
        dtype = self._kl_div(y_true, y_pred).dtype
        y_true = tf.cast(y_true, dtype)
        y_pred = tf.cast(y_pred, dtype)
        state = 0.5 * self._kl_div(
            y_true, 0.5 * (y_true + y_pred), sample_weight=sample_weight
        ) + 0.5 * self._kl_div(
            y_pred, 0.5 * (y_true + y_pred), sample_weight=sample_weight
        )
        self._metric_values.assign(state)



import tensorflow as tf
from tensorflow.keras.metrics import MeanSquaredError as TF_MSE

from calotron.metrics.BaseMetric import BaseMetric


class MeanSquaredError(BaseMetric):
    def __init__(self, name="mse", dtype=None, **kwargs):
        super().__init__(name, dtype, **kwargs)
        self._mse = TF_MSE(name=name, dtype=dtype)

    def update_state(self, y_true, y_pred, sample_weight=None):
        state = self._mse(y_true, y_pred, sample_weight=sample_weight)
        self._metric_values.assign(state)



import tensorflow as tf

from calotron.layers import Decoder, Encoder, MultiActivations


class Transformer(tf.keras.Model):
    def __init__(
        self,
        output_depth,
        encoder_depth,
        decoder_depth,
        num_layers,
        num_heads,
        key_dim=None,
        encoder_pos_dim=None,
        decoder_pos_dim=None,
        encoder_pos_normalization=128,
        decoder_pos_normalization=128,
        encoder_max_length=32,
        decoder_max_length=32,
        ff_units=256,
        dropout_rate=0.1,
        pos_sensitive=False,
        residual_smoothing=True,
        output_activations=None,
        name=None,
        dtype=None,
    ):
        super().__init__(name=name, dtype=dtype)
        self._output_depth = int(output_depth)
        self._encoder_depth = int(encoder_depth)
        self._decoder_depth = int(decoder_depth)
        self._num_layers = int(num_layers)
        self._num_heads = int(num_heads)
        self._key_dim = int(key_dim) if key_dim else None
        self._pos_dim = (
            int(encoder_pos_dim) if encoder_pos_dim else None,
            int(decoder_pos_dim) if decoder_pos_dim else None,
        )
        self._pos_normalization = (
            float(encoder_pos_normalization),
            float(decoder_pos_normalization),
        )
        self._max_length = (int(encoder_max_length), int(decoder_max_length))
        self._ff_units = int(ff_units)
        self._dropout_rate = float(dropout_rate)
        self._pos_sensitive = bool(pos_sensitive)
        self._residual_smoothing = bool(residual_smoothing)

        self._encoder = Encoder(
            encoder_depth=self._encoder_depth,
            num_layers=self._num_layers,
            num_heads=self._num_heads,
            key_dim=self._key_dim,
            pos_dim=self._pos_dim[0],
            pos_normalization=self._pos_normalization[0],
            max_length=self._max_length[0],
            ff_units=self._ff_units,
            dropout_rate=self._dropout_rate,
            pos_sensitive=self._pos_sensitive,
            residual_smoothing=self._residual_smoothing,
            dtype=self.dtype,
        )

        self._decoder = Decoder(
            decoder_depth=self._decoder_depth,
            num_layers=self._num_layers,
            num_heads=self._num_heads,
            key_dim=self._key_dim,
            pos_dim=self._pos_dim[1],
            pos_normalization=self._pos_normalization[1],
            max_length=self._max_length[1],
            ff_units=self._ff_units,
            dropout_rate=self._dropout_rate,
            pos_sensitive=self._pos_sensitive,
            residual_smoothing=self._residual_smoothing,
            dtype=self.dtype,
        )

        self._final_layer = tf.keras.layers.Dense(
            self._output_depth, name="output_layer", dtype=self.dtype
        )

        if output_activations is not None:
            # TODO: find a way to remove the whole list of activations from the summary
            self._multi_act_layer = MultiActivations(
                output_activations,
                self._output_depth,
                name="ma_layer",
                dtype=self.dtype,
            )
            self._output_activations = self._multi_act_layer.output_activations
        else:
            self._output_activations = None

    def call(self, inputs):
        source, target = inputs
        context = self._encoder(
            x=source
        )  # shape: (batch_size, source_elements, encoder_depth)
        output = self._decoder(
            x=target, context=context
        )  # shape: (batch_size, target_elements, decoder_depth)
        output = self._final_layer(
            output
        )  # shape: (batch_size, target_elements, output_depth)
        if self._output_activations is not None:
            output = self._multi_act_layer(
                output
            )  # shape: (batch_size, target_elements, output_depth)
        return output

    @property
    def output_depth(self) -> int:
        return self._output_depth

    @property
    def encoder_depth(self) -> int:
        return self._encoder_depth

    @property
    def decoder_depth(self) -> int:
        return self._decoder_depth

    @property
    def num_layers(self) -> int:
        return self._num_layers

    @property
    def num_heads(self) -> int:
        return self._num_heads

    @property
    def key_dim(self):  # TODO: add Union[int, None]
        return self._key_dim

    @property
    def pos_dim(self):  # TODO: add Union[tuple, None]
        return self._pos_dim

    @property
    def pos_normalization(self) -> tuple:
        return self._pos_normalization

    @property
    def max_length(self) -> tuple:
        return self._max_length

    @property
    def ff_units(self) -> int:
        return self._ff_units

    @property
    def dropout_rate(self) -> float:
        return self._dropout_rate

    @property
    def pos_sensitive(self) -> bool:
        return self._pos_sensitive

    @property
    def residual_smoothing(self) -> bool:
        return self._residual_smoothing

    @property
    def output_activations(self):  # TODO: add Union[list, None]
        return self._output_activations

    @property
    def encoder(self) -> Encoder:
        return self._encoder

    @property
    def decoder(self) -> Decoder:
        return self._decoder



import tensorflow as tf
from tensorflow.keras.metrics import KLDivergence as TF_KLDivergence

from calotron.metrics.BaseMetric import BaseMetric


class KLDivergence(BaseMetric):
    def __init__(self, name="kl_div", dtype=None):
        super().__init__(name, dtype)
        self._kl_div = TF_KLDivergence(name=name, dtype=dtype)

    def update_state(self, y_true, y_pred, sample_weight=None):
        state = self._kl_div(y_true, y_pred, sample_weight=sample_weight)
        self._metric_values.assign(state)



import tensorflow as tf
from tensorflow.keras.losses import KLDivergence as TF_KLDivergence

from calotron.losses.BaseLoss import BaseLoss


class JSDivergence(BaseLoss):
    def __init__(self, reduction="auto", name="js_loss"):
        super().__init__(name)
        self._kl_div = TF_KLDivergence(reduction=reduction)

    def discriminator_loss(
        self, discriminator, target_true, target_pred, sample_weight=None
    ):
        y_true = discriminator(target_true)
        y_pred = discriminator(target_pred)
        return -self._js_div(
            y_true, y_pred, sample_weight=sample_weight
        )  # divergence maximization

    def transformer_loss(
        self, discriminator, target_true, target_pred, sample_weight=None
    ):
        y_true = discriminator(target_true)
        y_pred = discriminator(target_pred)
        return self._js_div(
            y_true, y_pred, sample_weight=sample_weight
        )  # divergence minimization

    def _js_div(self, y_true, y_pred, sample_weight=None):
        dtype = self._kl_div(y_true, y_pred).dtype
        y_true = tf.cast(y_true, dtype=dtype)
        y_pred = tf.cast(y_pred, dtype=dtype)
        loss = 0.5 * self._kl_div(
            y_true, 0.5 * (y_true + y_pred), sample_weight=sample_weight
        ) + 0.5 * self._kl_div(
            y_pred, 0.5 * (y_true + y_pred), sample_weight=sample_weight
        )
        return loss



import tensorflow as tf

from calotron.models.Discriminator import Discriminator
from calotron.models.Transformer import Transformer
from calotron.utils import checkLoss, checkMetrics, checkOptimizer


class Calotron(tf.keras.Model):
    def __init__(self, transformer, discriminator):
        super().__init__()
        if not isinstance(transformer, Transformer):
            raise TypeError(
                f"`transformer` should be a calotron's "
                f"`Transformer`, instead "
                f"{type(transformer)} passed"
            )
        self._transformer = transformer
        if not isinstance(discriminator, Discriminator):
            raise TypeError(
                f"`discriminator` should be a calotron's "
                f"`Discriminator`, instead "
                f"{type(discriminator)} passed"
            )
        self._discriminator = discriminator

    def call(self, inputs):
        source, target = inputs
        output = self._transformer((source, target))
        d_output_true = self._discriminator(target)
        d_output_pred = self._discriminator(output)
        return output, d_output_true, d_output_pred

    def summary(self, **kwargs):
        print("_" * 65)
        self._transformer.summary(**kwargs)
        self._discriminator.summary(**kwargs)

    def compile(
        self,
        loss,
        metrics=None,
        transformer_optimizer="rmsprop",
        discriminator_optimizer="rmsprop",
        transformer_upds_per_batch=1,
        discriminator_upds_per_batch=1,
    ):
        super().compile()
        self._loss = checkLoss(loss)
        self._t_loss = tf.keras.metrics.Mean(name=f"t_{self._loss.name}")
        self._d_loss = tf.keras.metrics.Mean(name=f"d_{self._loss.name}")
        self._metrics = checkMetrics(metrics)
        self._t_opt = checkOptimizer(transformer_optimizer)
        self._d_opt = checkOptimizer(discriminator_optimizer)
        if transformer_upds_per_batch < 1:
            raise ValueError("`transformer_upds_per_batch` should be greater than 0")
        self._t_upds_per_batch = int(transformer_upds_per_batch)
        if discriminator_upds_per_batch < 1:
            raise ValueError("`discriminator_upds_per_batch` should be greater than 0")
        self._d_upds_per_batch = int(discriminator_upds_per_batch)

    def train_step(self, data):
        if len(data) == 3:
            source, target, sample_weight = data
        else:
            source, target = data
            sample_weight = None
        target_in, target_out = self._prepare_target(target)

        for _ in range(self._d_upds_per_batch):
            self._d_train_step(source, target_in, target_out, sample_weight)
        for _ in range(self._t_upds_per_batch):
            self._t_train_step(source, target_in, target_out, sample_weight)

        train_dict = dict()
        if self._metrics is not None:
            for metric in self._metrics:
                train_dict.update({metric.name: metric.result()})
        train_dict.update(
            {
                f"t_{self._loss.name}": self._t_loss.result(),
                f"d_{self._loss.name}": self._d_loss.result(),
                "t_lr": self._t_opt.learning_rate,
                "d_lr": self._d_opt.learning_rate,
            }
        )
        return train_dict

    @staticmethod
    def _prepare_target(target):
        start_token = tf.reduce_mean(target, axis=(0, 1))[None, None, :]
        start_token = tf.tile(start_token, (target.shape[0], 1, 1))
        target_in = tf.concat([start_token, target[:, :-1, :]], axis=1)
        target_out = target
        return target_in, target_out

    def _d_train_step(self, source, target_in, target_out, sample_weight):
        with tf.GradientTape() as tape:
            target_pred = self._transformer((source, target_in))
            loss = self._loss.discriminator_loss(
                self._discriminator,
                target_out,
                target_pred,
                sample_weight=sample_weight,
            )
        grads = tape.gradient(loss, self._discriminator.trainable_weights)
        self._d_opt.apply_gradients(zip(grads, self._discriminator.trainable_weights))
        self._d_loss.update_state(loss)

    def _t_train_step(self, source, target_in, target_out, sample_weight):
        with tf.GradientTape() as tape:
            target_pred = self._transformer((source, target_in))
            loss = self._loss.transformer_loss(
                self._discriminator,
                target_out,
                target_pred,
                sample_weight=sample_weight,
            )
        grads = tape.gradient(loss, self._transformer.trainable_weights)
        self._t_opt.apply_gradients(zip(grads, self._transformer.trainable_weights))
        self._t_loss.update_state(loss)
        if self._metrics is not None:
            y_pred = self._discriminator(target_pred)
            y_true = self._discriminator(target_out)
            for metric in self._metrics:
                metric.update_state(y_true, y_pred)

    @staticmethod
    def get_start_token(target):
        start_token = tf.reduce_mean(target, axis=(0, 1))[None, :]
        start_token = tf.tile(start_token, (target.shape[0], 1))
        return start_token

    @property
    def transformer(self) -> Transformer:
        return self._transformer

    @property
    def discriminator(self) -> Discriminator:
        return self._discriminator

    @property
    def metrics(self) -> list:
        reset_states = [self._t_loss, self._d_loss]
        if self._metrics is not None:
            reset_states += self._metrics
        return reset_states

    @property
    def transformer_optimizer(self) -> tf.keras.optimizers.Optimizer:
        return self._t_opt

    @property
    def discriminator_optimizer(self) -> tf.keras.optimizers.Optimizer:
        return self._d_opt

    @property
    def transformer_upds_per_batch(self) -> int:
        return self._t_upds_per_batch

    @property
    def discriminator_upds_per_batch(self) -> int:
        return self._d_upds_per_batch



import tensorflow as tf

from calotron.models import Transformer


class Simulator(tf.Module):
    def __init__(self, transformer, start_token):
        super().__init__()
        if not isinstance(transformer, Transformer):
            raise TypeError(
                f"`transformer` should be a calotron's "
                f"`{type(Transformer).__name__}`, instead "
                f"{type(transformer)} passed"
            )
        self._transformer = transformer
        if not isinstance(start_token, tf.Tensor):
            raise TypeError(
                f"`start_token` should be a TensorFlow "
                f"`Tensor`, instead {type(start_token)} passed"
            )
        self._start_token = start_token

    def __call__(self, source, max_length):
        if not isinstance(source, tf.Tensor):
            raise TypeError(
                f"`source` should be a TensorFlow "
                f"`Tensor`, instead {type(source)} passed"
            )
        if max_length < 1:
            raise ValueError("`max_length` should be greater than 0")
        max_length = int(max_length)

        start_token = tf.cast(self._start_token, dtype=source.dtype)
        target = tf.expand_dims(start_token, axis=1)
        for _ in tf.range(max_length):
            predictions = self.transformer((source, target), training=False)
            target = tf.concat([target, predictions[:, -1:, :]], axis=1)

        assert target.shape[1] == max_length + 1
        return target[:, 1:, :]

    @property
    def transformer(self) -> Transformer:
        return self._transformer

    @property
    def start_token(self) -> tf.Tensor:
        return self._start_token



import tensorflow as tf


class BaseAttention(tf.keras.layers.Layer):
    def __init__(self, **kwargs):
        super().__init__()
        self._mha = tf.keras.layers.MultiHeadAttention(**kwargs)
        self._layernorm = tf.keras.layers.LayerNormalization()
        self._add = tf.keras.layers.Add()


class CrossAttention(BaseAttention):
    def call(self, x, context):  # x -> target, context -> source
        attn_output = self._mha(query=x, key=context, value=context)
        x = self._add([x, attn_output])
        x = self._layernorm(x)
        return x  # shape: (batch_size, x_elements, x_depth)


class GlobalSelfAttention(BaseAttention):
    def call(self, x):  # x -> source
        attn_output = self._mha(query=x, key=x, value=x)
        x = self._add([x, attn_output])
        x = self._layernorm(x)
        return x  # shape: (batch_size, x_elements, x_depth)


class CausalSelfAttention(BaseAttention):
    def call(self, x):  # x -> target
        attn_output = self._mha(query=x, key=x, value=x, use_causal_mask=True)
        x = self._add([x, attn_output])
        x = self._layernorm(x)
        return x  # shape: (batch_size, x_elements, x_depth)



import tensorflow as tf

from calotron.losses import BinaryCrossentropy as BCE
from calotron.losses import JSDivergence as JS
from calotron.losses import KLDivergence as KL
from calotron.losses import MeanAbsoluteError as MAE
from calotron.losses import MeanSquaredError as MSE
from calotron.losses.BaseLoss import BaseLoss

LOSS_SHORTCUTS = ["bce", "kl", "js", "mse", "mae"]
CALOTRON_LOSSES = [BCE(), KL(), JS(), MSE(), MAE()]


def checkLoss(loss) -> BaseLoss:
    if isinstance(loss, str):
        if loss in LOSS_SHORTCUTS:
            for str_loss, calo_loss in zip(LOSS_SHORTCUTS, CALOTRON_LOSSES):
                if loss == str_loss:
                    return calo_loss
        else:
            raise ValueError(
                f"`loss` should be selected in {LOSS_SHORTCUTS}, "
                f"instead '{loss}' passed"
            )
    elif isinstance(loss, BaseLoss):
        return loss
    else:
        raise TypeError(
            f"`loss` should be a calotron's `BaseLoss`, " f"instead {type(loss)} passed"
        )



import tensorflow as tf


class BaseMetric(tf.keras.metrics.Metric):
    def __init__(self, name="metric", dtype=None):
        super().__init__(name, dtype)
        self._metric_values = self.add_weight(
            name=f"{name}_values", initializer="zeros"
        )

    def update_state(self, y_true, y_pred, sample_weight=None):
        raise NotImplementedError(
            "Only `BaseMetric` subclasses have the "
            "`update_state()` method implemented."
        )

    def result(self):
        return self._metric_values



import tensorflow as tf
from tensorflow.keras.metrics import BinaryCrossentropy as TF_BCE

from calotron.metrics.BaseMetric import BaseMetric


class BinaryCrossentropy(BaseMetric):
    def __init__(self, name="bce", dtype=None, from_logits=False, label_smoothing=0.0):
        super().__init__(name, dtype)
        self._bce = TF_BCE(
            name=name,
            dtype=dtype,
            from_logits=from_logits,
            label_smoothing=label_smoothing,
        )

    def update_state(self, y_true, y_pred, sample_weight=None):
        state = self._bce(tf.ones_like(y_pred), y_pred, sample_weight=sample_weight)
        self._metric_values.assign(state)



import tensorflow as tf
from tensorflow.keras.layers import Activation, Layer


def checkActivations(activations, output_length, dtype=None):
    if activations is None:
        return None
    elif isinstance(activations, str):
        return [Activation(activations, dtype=dtype) for _ in range(output_length)]
    elif isinstance(activations, Layer):
        return [activations for _ in range(output_length)]
    elif isinstance(activations, list):
        if len(activations) != output_length:
            raise ValueError(
                f"`activations` list should have "
                f"{output_length} elements, instead "
                f"{len(activations)} passed"
            )
        checked_activations = list()
        for activation in activations:
            if isinstance(activation, str):
                checked_activations.append(Activation(activation, dtype=dtype))
            elif isinstance(activation, Layer):
                checked_activations.append(activation)
            else:
                raise ValueError(
                    f"`activations` elements should be strings "
                    f"or TensorFlow `Activation` layers, "
                    f"instead {type(activation)} passed"
                )
        return checked_activations
    else:
        raise TypeError(
            f"`activations` should be a string, a TensorFlow "
            f"`Activation` layer, or a list of such objects, "
            f"instead {type(activations)} passed"
        )



from calotron.metrics import Accuracy
from calotron.metrics import BinaryCrossentropy as BCE
from calotron.metrics import JSDivergence as JS
from calotron.metrics import KLDivergence as KL
from calotron.metrics import MeanAbsoluteError as MAE
from calotron.metrics import MeanSquaredError as MSE
from calotron.metrics import RootMeanSquaredError as RMSE
from calotron.metrics.BaseMetric import BaseMetric

METRIC_SHORTCUTS = ["accuracy", "bce", "kl_div", "js_div", "mse", "rmse", "mae"]
CALOTRON_METRICS = [Accuracy(), BCE(), KL(), JS(), MSE(), RMSE(), MAE()]


def checkMetrics(metrics):
    if metrics is None:
        return None
    else:
        if isinstance(metrics, list):
            checked_metrics = list()
            for metric in metrics:
                if isinstance(metric, str):
                    if metric in METRIC_SHORTCUTS:
                        for str_metric, calo_metric in zip(
                            METRIC_SHORTCUTS, CALOTRON_METRICS
                        ):
                            if metric == str_metric:
                                checked_metrics.append(calo_metric)
                    else:
                        raise ValueError(
                            f"`metrics` elements should be selected in "
                            f"{METRIC_SHORTCUTS}, instead '{metric}' passed"
                        )
                elif isinstance(metric, BaseMetric):
                    checked_metrics.append(metric)
                else:
                    raise TypeError(
                        f"`metrics` elements should be a calotron's "
                        f"`BaseMetric`, instead {type(metric)} passed"
                    )
            return checked_metrics
        else:
            raise TypeError(
                f"`metrics` should be a list of strings or calotron's "
                f"`BaseMetric`s, instead {type(metrics)} passed"
            )



class BaseLoss:
    def __init__(self, name="loss"):
        if not isinstance(name, str):
            raise TypeError(
                f"`name` should be a string " f"instead {type(name)} passed"
            )
        self._name = name
        self._loss = None

    def discriminator_loss(
        self, discriminator, target_true, target_pred, sample_weight=None
    ):
        raise NotImplementedError(
            "Only `BaseLoss` subclasses have the "
            "`discriminator_loss()` method implemented."
        )

    def transformer_loss(
        self, discriminator, target_true, target_pred, sample_weight=None
    ):
        raise NotImplementedError(
            "Only `BaseLoss` subclasses have the "
            "`transformer_loss()` method implemented."
        )

    @property
    def name(self) -> str:
        return self._name



import numpy as np
import tensorflow as tf
from tensorflow.keras.callbacks import Callback

K = tf.keras.backend


class BaseScheduler(Callback):
    def __init__(self, optimizer, verbose=False):
        super().__init__()
        self._optimizer = optimizer
        self._verbose = bool(verbose)

    def on_train_begin(self, logs=None):
        init_lr = K.get_value(self._optimizer.learning_rate)
        self._init_lr = tf.identity(init_lr)
        self._dtype = self._init_lr.dtype
        self._step = -1

    def on_batch_begin(self, batch, logs=None):
        self._step += 1
        step = tf.cast(self._step, self._dtype)
        K.set_value(
            self._optimizer.learning_rate, self._scheduled_lr(self._init_lr, step)
        )

    def _scheduled_lr(self, init_lr, step):
        return init_lr

    def on_batch_end(self, batch, logs=None):
        logs = logs or {}
        if self._verbose:
            logs["lr"] = K.get_value(self.model.optimizer.learning_rate)

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        if self._verbose:
            logs["lr"] = K.get_value(self.model.optimizer.learning_rate)

    @property
    def optimizer(self) -> tf.keras.optimizers.Optimizer:
        return self._optimizer


class CosineDecay(BaseScheduler):
    def __init__(self, optimizer, decay_steps, alpha=0.0, verbose=False):
        super().__init__(optimizer, verbose)
        assert decay_steps > 0
        self._decay_steps = int(decay_steps)
        assert (alpha) >= 0.0 and (alpha <= 1.0)
        self._alpha = float(alpha)

    def on_train_begin(self, logs=None):
        super().on_train_begin(logs=logs)
        self._tf_decay_steps = tf.cast(self._decay_steps, self._dtype)
        self._tf_alpha = tf.cast(self._alpha, self._dtype)

    def _scheduled_lr(self, init_lr, step):
        step = tf.minimum(step, self._tf_decay_steps)
        p = tf.divide(step, self._tf_decay_steps)
        cosine_decay = 0.5 * (1 + tf.cos(tf.constant(np.pi) * p))
        decayed = tf.multiply(1 - self._tf_alpha, cosine_decay + self._tf_alpha)
        return tf.multiply(init_lr, decayed)

    @property
    def decay_steps(self) -> int:
        return self._decay_steps

    @property
    def alpha(self) -> float:
        return self._alpha


class ExponentialDecay(BaseScheduler):
    def __init__(
        self, optimizer, decay_rate, decay_steps, staircase=False, verbose=False
    ):
        super().__init__(optimizer, verbose)
        assert decay_rate > 0.0
        self._decay_rate = float(decay_rate)
        assert decay_steps > 0
        self._decay_steps = int(decay_steps)
        assert isinstance(staircase, bool)
        self._staircase = staircase

    def on_train_begin(self, logs=None):
        super().on_train_begin(logs=logs)
        self._tf_decay_rate = tf.cast(self._decay_rate, self._dtype)
        self._tf_decay_steps = tf.cast(self._decay_steps, self._dtype)

    def _scheduled_lr(self, init_lr, step):
        p = tf.divide(step, self._tf_decay_steps)
        if self._staircase:
            p = tf.floor(p)
        return tf.multiply(init_lr, tf.pow(self._tf_decay_rate, p))

    @property
    def decay_rate(self) -> float:
        return self._decay_rate

    @property
    def decay_steps(self) -> int:
        return self._decay_steps

    @property
    def staircase(self) -> bool:
        return self._staircase


class InverseTimeDecay(BaseScheduler):
    def __init__(
        self, optimizer, decay_rate, decay_steps, staircase=False, verbose=False
    ):
        super().__init__(optimizer, verbose)
        assert decay_rate > 0.0
        self._decay_rate = float(decay_rate)
        assert decay_steps > 0
        self._decay_steps = int(decay_steps)
        assert isinstance(staircase, bool)
        self._staircase = staircase

    def on_train_begin(self, logs=None):
        super().on_train_begin(logs=logs)
        self._tf_decay_rate = tf.cast(self._decay_rate, self._dtype)
        self._tf_decay_steps = tf.cast(self._decay_steps, self._dtype)

    def _scheduled_lr(self, init_lr, step):
        p = tf.divide(step, self._tf_decay_steps)
        if self._staircase:
            p = tf.floor(p)
        return tf.divide(init_lr, 1 + tf.multiply(self._tf_decay_rate, p))

    @property
    def decay_rate(self) -> float:
        return self._decay_rate

    @property
    def decay_steps(self) -> int:
        return self._decay_steps

    @property
    def staircase(self) -> bool:
        return self._staircase


class PiecewiseConstantDecay(BaseScheduler):
    def __init__(self, optimizer, boundaries, values, verbose=False):
        super().__init__(optimizer, verbose)
        assert isinstance(boundaries, (list, tuple, np.ndarray))
        assert isinstance(values, (list, tuple, np.ndarray))
        assert len(boundaries) >= 1
        assert len(values) >= 2
        assert len(boundaries) == len(values) - 1
        self._boundaries = [0] + [int(b) for b in boundaries]
        self._values = [float(v) for v in values]

    def on_train_begin(self, logs=None):
        super().on_train_begin(logs=logs)
        self._tf_boundaries = tf.cast(self._boundaries, self._dtype)
        self._tf_values = tf.cast(self._values, self._dtype)

    def _scheduled_lr(self, init_lr, step):
        for i in range(len(self._boundaries) - 1):
            if (step >= self._tf_boundaries[i]) and (step < self._tf_boundaries[i + 1]):
                return self._tf_values[i]
        return self._tf_values[-1]

    @property
    def boundaries(self) -> list:
        return self._boundaries

    @property
    def values(self) -> list:
        return self._values


class PolynomialDecay(BaseScheduler):
    def __init__(
        self,
        optimizer,
        decay_steps,
        end_learning_rate=0.0001,
        power=1.0,
        cycle=False,
        verbose=False,
    ):
        super().__init__(optimizer, verbose)
        assert decay_steps > 0
        self._decay_steps = int(decay_steps)
        assert end_learning_rate > 0.0
        self._end_learning_rate = float(end_learning_rate)
        assert power > 0.0
        self._power = float(power)
        assert isinstance(cycle, bool)
        self._cycle = cycle

    def on_train_begin(self, logs=None):
        super().on_train_begin(logs=logs)
        self._tf_decay_steps = tf.cast(self._decay_steps, self._dtype)
        self._tf_end_learning_rate = tf.cast(self._end_learning_rate, self._dtype)
        self._tf_power = tf.cast(self._power, self._dtype)

    def _scheduled_lr(self, init_lr, step):
        if not self._cycle:
            step = tf.minimum(step, self._tf_decay_steps)
            decay_steps = self._tf_decay_steps
        else:
            decay_steps = tf.multiply(
                self._tf_decay_steps,
                tf.math.ceil(tf.divide(step, self._tf_decay_steps)),
            )
        return (
            (init_lr - self._tf_end_learning_rate)
            * tf.pow(1 - step / decay_steps, self._tf_power)
        ) + self._tf_end_learning_rate

    @property
    def decay_steps(self) -> int:
        return self._decay_steps

    @property
    def end_learning_rate(self) -> float:
        return self._end_learning_rate

    @property
    def power(self) -> float:
        return self._power

    @property
    def cycle(self) -> bool:
        return self._cycle


class AttentionDecay(BaseScheduler):
    def __init__(self, optimizer, d_model, warmup_steps=4000, verbose=False):
        super().__init__(optimizer, verbose)
        assert d_model > 0
        self._d_model = int(d_model)
        assert warmup_steps > 0
        self._warmup_steps = int(warmup_steps)

    def on_train_begin(self, logs=None):
        super().on_train_begin(logs=logs)
        self._tf_d_model = tf.cast(self._d_model, self._dtype)
        self._tf_warmup_steps = tf.cast(self._warmup_steps, self._dtype)

    def _scheduled_lr(self, init_lr, step):
        arg1 = tf.math.rsqrt(step)
        arg2 = tf.multiply(step, tf.pow(self._tf_warmup_steps, -1.5))
        return tf.multiply(tf.math.rsqrt(self._tf_d_model), tf.math.minimum(arg1, arg2))

    @property
    def d_model(self) -> int:
        return self._d_model

    @property
    def warmup_steps(self) -> int:
        return self._warmup_steps



import tensorflow as tf
from tensorflow.keras.optimizers import SGD, Adam, Optimizer, RMSprop

OPT_SHORTCUTS = ["sgd", "rmsprop", "adam"]
TF_OPTIMIZERS = [SGD(), RMSprop(), Adam()]


def checkOptimizer(optimizer) -> Optimizer:
    if isinstance(optimizer, str):
        if optimizer in OPT_SHORTCUTS:
            for opt, tf_opt in zip(OPT_SHORTCUTS, TF_OPTIMIZERS):
                if optimizer == opt:
                    return tf_opt
        else:
            raise ValueError(
                f"`optimizer` should be selected in {OPT_SHORTCUTS}, "
                f"instead '{optimizer}' passed"
            )
    elif isinstance(optimizer, Optimizer):
        return optimizer
    else:
        raise TypeError(
            f"`optimizer` should be a TensorFlow `Optimizer`, "
            f"instead {type(optimizer)} passed"
        )



from .Attention import CausalSelfAttention, CrossAttention, GlobalSelfAttention
from .Decoder import Decoder
from .DeepSets import DeepSets
from .Encoder import Encoder
from .FeedForward import FeedForward
from .MultiActivations import MultiActivations
from .PositionalEmbedding import PositionalEmbedding
